import json
from typing import List, Dict, Tuple, NewType, Optional
from functools import reduce

from elasticsearch import Elasticsearch as Es
from elasticsearch.exceptions import RequestError


import logging

logger = logging.getLogger(__name__)

VectorDict = NewType('VectorDict', Dict[str, List[float]])
ErrorString = Optional[str]


class IndexAlreadyExists(Exception):
    def __init__(self, index_name, *args, **kwargs):
        return super().__init__(f'The index "{index_name}" already exists')


# Filter response data attributes
FILTER_PATH = ['hits.hits._score', 'hits.hits._id', 'hits.hits._source', 'hits.total', 'hits.max_score']


def create_data(doc_id: int, vectors: Dict[str, List[float]], index_name: str):
    data: List[Dict] = []
    data.append(dict(index=dict(_index=index_name, _id=doc_id)))
    data.append({vector_name: vector for vector_name, vector in vectors.items()})
    return data


def es_bulk(index_name: str, data: List[Dict], es: Es) -> Tuple[bool, str]:
    response = es.bulk(index=index_name, body=data)
    if response['errors']:
        return False, response['items'] and response['items'][0]['index']['error']['caused_by']['reason']
    return True, 'Success'


def add_to_index(doc_id: int, vectors: VectorDict, index_name: str, es: Es):
    data = create_data(doc_id, vectors, index_name)
    return es_bulk(index_name, data, es)


def add_to_index_bulk(doc_ids: List[int], vectors: List[VectorDict], index_name: str, es: Es) -> Tuple[bool, str]:
    # Check if lengths match
    if len(doc_ids) != len(vectors):
        return False, "Mismatching number of doc_ids and vectors"
    data: List[Dict] = reduce(
        lambda acc, id_vector: [*acc, *create_data(id_vector[0], id_vector[1], index_name)],
        zip(doc_ids, vectors),
        []
    )
    return es_bulk(index_name, data, es)


def search_similar(similar_count: int, vector: Tuple[str, List[float]], index_name: str, es: Es):
    query = {
        "size": similar_count,
        "query": {
            "knn": {
                vector[0]: {
                    "vector": vector[1],
                    "k": similar_count
                }
            }
        }
    } if vector else None
    return es.search(body=query, index=index_name, filter_path=FILTER_PATH)


def index_exists(index_name: str, es: Es) -> bool:
    """
    Checks if index named `index_name` exists
    """
    return es.indices.exists(index_name)


def create_knn_vector_index_if_not_exists(
        index_name: str, vector_size: int, es: Es
        ) -> Tuple[bool, ErrorString]:
    return create_knn_vector_index(index_name, vector_size, es, ignore_error=True)


def create_knn_vector_index(
        index_name: str, vector_size: int, es: Es, ignore_error: bool = False
        ) -> Tuple[bool, ErrorString]:
    """
    Create knn vector index with given `index_name` and `vector_size`.
    The vector name will be `vector1`
    """
    if es.indices.exists(index_name):
        if ignore_error:
            return True, None
        return False, f"The index '{index_name}' already exists"

    properties = dict(
        vector1=dict(
            type="knn_vector",
            dimension=vector_size
        )
    )
    data = dict(
        settings={"index": {"knn": True, "knn.space_type": "cosinesimil"}},
        mappings=dict(
            properties=properties
        )
    )

    try:
        es.indices.create(index_name, body=data)
        return True, None
    except RequestError as e:
        if e.args[1] == 'resource_already_exists_exception':
            return False, f"The index '{index_name}' already exists"
        reason = e.args[2]['error']['root_cause'][0]['reason']
        logger.warning(reason)
        return False, reason



import sys
import string
import json
import numpy as np

RAW_TRIGRAMS_PATH = 'trigrams_statistics_raw.json'
DOCUMENTS_SUMMARY_PATH = 'documents_summary.json'

ENGLISH = 'en'
SPANISH = 'es'
FRENCH = 'fr'

LANGS = [ENGLISH, SPANISH, FRENCH]

ENGLISH_ALPHABET = set(string.ascii_lowercase + string.digits + ' ')
SPANISH_ALPHABET = {*ENGLISH_ALPHABET, 'ñ'}
FRENCH_ALPHABET = {*ENGLISH_ALPHABET, *set('çéâêîôûàèìòùëïü')}


def all_digits(text) -> bool:
    text = text.strip()  # strip spaces
    for x in text:
        if x not in string.digits:
            return False
    return True


def all_nondigits(text) -> bool:
    for x in text:
        if x in string.digits:
            return False
    return True


def filter_non_alaphabetic_trigrams(all_trigrams):
    filtered_trigrams = {}
    filtered_trigrams[ENGLISH] = {
        k: v
        for k, v in all_trigrams[ENGLISH].items()
        if all(x in ENGLISH_ALPHABET for x in k) and (all_digits(k) or all_nondigits(k))  # Do not allow things like '2ba' or 'a2d' but allow ' 34' or '22 '
    }
    filtered_trigrams[SPANISH] = {
        k: v
        for k, v in all_trigrams[SPANISH].items()
        if all(x in SPANISH_ALPHABET for x in k) and (all_digits(k) or all_nondigits(k))
    }
    filtered_trigrams[FRENCH] = {
        k: v
        for k, v in all_trigrams[FRENCH].items()
        if all(x in FRENCH_ALPHABET for x in k) and (all_digits(k) or all_nondigits(k))
    }
    return filtered_trigrams


def normalize_all_trigrams(all_trigrams):
    totals = {
        lang: sum(v for v in all_trigrams[lang].values())
        for lang in LANGS
    }
    return {
        lang: {k: v / totals[lang] for k, v in all_trigrams[lang].items()}
        for lang in LANGS
    }


# NOT USED
def filter_significant_trigrams(all_trigrams, n=80):
    """ This calculates the standard deviation and returns values within n * sigma"""
    langs = [ENGLISH, SPANISH, FRENCH]
    means = {
        k: np.array(list(all_trigrams[k].values())).mean(dtype=np.float64)
        for k in langs
    }
    stds = {
        k: np.array(list(all_trigrams[k].values())).std(dtype=np.float64)
        for k in langs
    }
    print(means)
    print(stds)
    return {
        lang: {k: v for k, v in all_trigrams[lang].items() if v - means[lang] < -n * stds[lang]}
        for lang in langs
    }


def get_counts(all_trigrams):
    return {
        lang: len(all_trigrams[lang].keys())
        for lang in LANGS
    }


def write_to_file_key_vals(lang, data, prefix):
    with open(f'{lang}-{prefix}.csv', 'w') as f:
        for k, v in sorted(data.items(), key=lambda x: x[1], reverse=True):
            f.write(f'{k},{v}\n')


def generate_csvs(raw_path=RAW_TRIGRAMS_PATH):
    with open(raw_path) as f:
        all_trigrams = json.load(f)
        print('Raw trigrams count:')
        print(get_counts(all_trigrams))
        filtered_trigrams = filter_non_alaphabetic_trigrams(all_trigrams)
        [write_to_file_key_vals(lang, filtered_trigrams[lang], 'filtered') for lang in LANGS]
        normalized_trigrams = normalize_all_trigrams(filtered_trigrams)
        [write_to_file_key_vals(lang, normalized_trigrams[lang], 'normalized') for lang in LANGS]
        print('Filtered trigrams count:')
        print(get_counts(normalized_trigrams))
        print('DONE!!')


if __name__ == '__main__':
    args = sys.argv
    if len(args) != 2:
        print('Usage: python process_raw_trigrams.py <raw_json_path>')
        exit()
    path = args[1]
    generate_csvs(path)



import unittest

from .utils import remove_puncs_and_extra_spaces
from .vector_generator import (
    create_count_vector,
    normalize_count_vector,
    create_trigram_vector,
    EN_TRIGRAMS,
    ES_TRIGRAMS,
    FR_TRIGRAMS,
    InvalidText,
    InvalidLanguage,
    TrigramsNotLoaded,
)

VECTOR_SIZE = 10000


def test_remove_puncs_and_extra_spaces():
    pairs = [
        ('with.punctu!!ation$$', 'with punctu ation'),
        ('$!@#another punctuated te&&xt', 'another punctuated te xt'),
        ('multiple spaces   and  space at the end ', 'multiple spaces and space at the end'),
        ('Mix^^ of CA~`SEs!!', 'mix of ca ses'),
    ]
    for txt, processed in pairs:
        assert remove_puncs_and_extra_spaces(txt) == processed


def test_trigrams_count():
    assert len(EN_TRIGRAMS.keys()) == VECTOR_SIZE, f"{VECTOR_SIZE} trigrams for english"
    assert len(ES_TRIGRAMS.keys()) == VECTOR_SIZE, f"{VECTOR_SIZE} trigrams for spanish"
    assert len(FR_TRIGRAMS.keys()) == VECTOR_SIZE, f"{VECTOR_SIZE} trigrams for french"


class TestVectorCreation(unittest.TestCase):
    """Tests for vector generation"""

    def test_count_vector(self):
        custom_trigrams = {' ab': 0, 'abc': 1, ' bc': 2, 'bc ': 3, ' is': 4, 'is ': 5}
        text = 'Abc is abc. is ab'
        # positions for [' ab', 'abc', ' bc', 'bc ', ' is', 'is ']
        expected_count_vector = [2, 2, 0, 2, 2, 2]
        processed_text = remove_puncs_and_extra_spaces(text)
        vector = create_count_vector(processed_text, custom_trigrams)
        assert len(vector) == len(custom_trigrams.values())
        assert expected_count_vector == vector

    def test_normalize_count_vector(self):
        vector = [1, 2, 3, 4, 5, 5]
        expected_normalized = [0.05, 0.1, 0.15, 0.2, 0.25, 0.25]
        assert normalize_count_vector(vector) == expected_normalized

    def test_create_trigram_vector_invalid_lang(self):
        with self.assertRaises(InvalidLanguage):
            create_trigram_vector('ch', 'test text. does not matter')

    def test_create_trigram_vector_small_document(self):
        with self.assertRaises(InvalidText):
            create_trigram_vector('en', 'te')

        with self.assertRaises(InvalidText):
            create_trigram_vector('fr', '   . ')

        with self.assertRaises(InvalidText):
            create_trigram_vector('es', None)

    def test_create_trigram_vector(self):
        # NOTE: it's infeasible to check exactly the vectors generated
        text = 'Document deduplication for Humanitarian organization!!'
        vector = create_trigram_vector('en', text)
        assert len(vector) == VECTOR_SIZE, f"There should be {VECTOR_SIZE} dimensions"

        assert all(x <= 1 for x in vector), "Every dimension should be <= 1"
        assert any(x > 0 for x in vector), "Some dimension should be > 0"

    def test_different_texts_vectors(self):
        text1 = '''# Major typhoon devastates areas of central Vietnam\n\nKuala Lumpur/Hanoi/Geneva, 29 October, 2020 – A major typhoon has devastated areas of central Vietnam, with dozens of people feared dead in tragic landslides in Quang Nam.\n\nThere are 53 people buried and feared dead in two landslides caused by the storm, according to Vietnam Government authorities and Red Cross teams have been working through the night to help with rescue efforts.\n\nThe storm has worsened floods in areas of central Vietnam and has caused fresh flooding in new areas, including the central highlands near the Laos border.\n\nMadam Nguyen Thi Xuan Thu, Viet Nam Red Cross Society President said: “We are heartbroken by more tragic loss of life as this typhoon has brought further misery and hardships to hundreds of thousands of people in central Vietnam. Around 89,000 homes have roofs blown off, with many destroyed by this storm.”\n\n“Red Cross relief teams are working non-stop to rescue people and provide critical relief as hundreds of thousands of lives have been turned upside down with so many homes and livelihoods devastated in this massive storm.”\n\nInfrastructure has been damaged including electricity and roads, with over 700 communities without power. More food crops and safe drinking water supplies have also been damaged or destroyed in regions of Vietnam already reeling from some of the worst flooding in decades.\n\nNguyen Hung Ha, Bangkok based Program Coordinator, International Federation of Red Cross and Red Crescent Societies, said: “This massive storm is another crippling blow to millions of people already struggling to cope with some of the most dangerous floods on record in central Vietnam.”\n\n“Relief teams are stretched to the limit due to these back-to back storms. We must redouble our efforts to get critical relief supplies, food, drinking water, tarpaulins and blankets to all those who need it in the coming days and weeks,” Mr Nguyen Hung Ha said.\n\nIn response to the existing flooding and impacts as a result of Typhoon Molave, the International Federation of Red Cross and Red Crescent Societies (IFRC) has launched an [Emergency Appeal](https://reliefweb.int/node/3683160) for 3.9 million Swiss Francs to fund relief and recovery efforts for an estimated 160,000 people. IFRC has already provided 500,000 Swiss francs to support local emergency efforts.\n\nIt is estimated that at least 150,000 people are at immediate risk of food shortages and hunger after thousands of hectares of crops have been destroyed, while over 2 million cattle and poultry are dead or swept away in some of the worst flooding in decades.\n\nFor more information or to arrange an interview, contact:\n\nIn Bangkok: Preeti Abraham, +66 61 412 3910, preeti.abraham@ifrc.org\n\nIn Kuala Lumpur: Antony Balmain, +60 12 230 8451, antony.balmain@ifrc.org\n\nIn Geneva: Matthew Cochrane, Mobile +41 79 251 80 39, matthew.cochrane@ifrc.org\n\nAbout IFRC\n\nIFRC is the world’s largest humanitarian network, comprising 192 National Red Cross and Red Crescent Societies working to save lives and promote dignity around the world.[www.ifrc.org](http://www.ifrc.org) \\- Facebook - Twitter - YouTube'''
        text2 = '''# Deadliest shipwreck of the year claims at least 140 lives\n\n*Saint-Louis, Senegal *– At least 140 people have drowned after a vessel carrying around 200 migrants sank off the Senegalese coast, the deadliest shipwreck recorded in 2020.\n\nAccording to media sources, the Senegalese and Spanish navies, and fishermen who were nearby, rescued 59 people and retrieved the remains of 20 others.\n\nThe International Organization for Migration (IOM) is deeply saddened by this recent tragedy, which follows four shipwrecks recorded in the Central Mediterranean last week and another in the UK Channel.\n\n“We call for unity between governments, partners and the international community to dismantle trafficking and smuggling networks that take advantage of desperate youth,” said Bakary Doumbia, IOM Senegal Chief of Mission.\n\n“It is also important that we advocate for enhanced legal channels to undermine the traffickers’ business model and prevent loss of life.”\n\nThe vessel reportedly left Mbour, a coastal town in western Senegal on Saturday (24/10) bound for the Canary Islands, according to local community members. The boat caught fire a few hours after departure and capsized near Saint-Louis, on Senegal’s northwest coast.\n\nThe Government of Senegal and IOM have arranged a mission to travel to Saint- Louis to assess the needs of survivors and provide immediate psychosocial assistance.\n\nThe number of departures from West Africa to the Canary Islands has significantly increased in recent weeks.\n\nIOM Senegal has been monitoring departures from the coast with the assistance of members of the community since the beginning of September. In September alone, [14 boats carrying 663 migrants](https://migration.iom.int/reports/senegal-%E2%80%94-monitoring- movements-canary-islands-%E2%80%94-movements-and-departures- senegal-1%E2%80%9430?close=true) left Senegal for the Canary Islands. Of these departures, 26 per cent were reported to have experienced an incident or shipwreck.\n\nAccording to the Spanish government, 11,006 arrivals have been recorded in the Canary Islands this year compared to 2,557 arrivals during the same period last year. This is still far below peaks seen in 2006 when over 32,000 people arrived.\n\nWith this tragic shipwreck, at least 414 people are known to have died along this route in 2020 according to IOM’s [Missing Migrants Project](https://missingmigrants.iom.int/?utm_source=Unknown+List&utm_campaign=d823a44265-EMAIL_CAMPAIGN_2020_10_29_09_00&utm_medium=email&utm_term=0_-d823a44265-), which recorded 210 fatalities there in all of 2019.\n\n*For more information please contact Aïssatou Sy at IOM's Regional Office for West and Central Africa, Tel: +221 77 479 21 41, Email: aisy@iom.int *'''
        vector1 = create_trigram_vector('en', text1)
        vector2 = create_trigram_vector('en', text2)
        cosine_similarity = sum([x*y for x, y in zip(vector1, vector2)])
        assert vector1 != vector2, "Two vectors should be different"
        assert cosine_similarity != 1, "Cosine similarity should not be 1"



from typing import List, Dict

from .utils import remove_puncs_and_extra_spaces
from .trigrams import en, es, fr

import logging

logger = logging.getLogger(__name__)

SUPPORTED_LANGUAGES = ['en', 'es', 'fr']

EN_TRIGRAMS = {t: i for i, t in enumerate(en.trigrams)}
ES_TRIGRAMS = {t: i for i, t in enumerate(es.trigrams)}
FR_TRIGRAMS = {t: i for i, t in enumerate(fr.trigrams)}


class InvalidText(Exception):
    pass


class InvalidLanguage(Exception):
    pass


class TrigramsNotLoaded(Exception):
    pass


def create_count_vector(processed_text: str, trigrams: Dict[str, int]) -> List[int]:
    vec_size = len(trigrams.keys())
    max_count = max(trigrams.values())

    assert max_count + 1 == vec_size, "Maximum index should be one less than vector size"
    raw_vector = [0] * vec_size  # initialize vector

    for i in range(len(processed_text) - 2):
        trigram = processed_text[i: i + 3]
        if trigrams.get(trigram) is not None:  # NOTE: explicit None check because 0 is also present
            raw_vector[trigrams[trigram]] += 1
    return raw_vector


def normalize_count_vector(count_vector: List[int]) -> List[float]:
    total_count = sum(count_vector)
    return [round(x / total_count, 5) for x in count_vector]


def create_trigram_vector(lang: str, text: str) -> List[float]:
    trigrams = None
    if lang not in SUPPORTED_LANGUAGES:
        raise InvalidLanguage(f'Invalid lanaguage "{lang}" to process trigrams')

    if lang == 'en':
        trigrams = EN_TRIGRAMS
    elif lang == 'es':
        trigrams = ES_TRIGRAMS
    elif lang == 'fr':
        trigrams = FR_TRIGRAMS

    if not trigrams:
        raise TrigramsNotLoaded(f'Trigrams for language "{lang}" are not loaded.')

    if text is None:
        raise InvalidText('No document')

    punc_removed = remove_puncs_and_extra_spaces(text)
    if len(punc_removed) < 3:
        raise InvalidText('Too small document')
    count_vector = create_count_vector(punc_removed, trigrams)
    return normalize_count_vector(count_vector)



from setuptools import setup, find_namespace_packages

setup(
    name='deep_utils.deduplication',
    version='1.0.8',
    description='Document deduplication package for DEEP',
    long_description_content_type='text/markdown',
    long_description=open('./README.md').read(),
    license="MIT",

    author='Bibek Pandey',
    author_email='bibek.pandey@togglecorp.com',
    url="https://github.com/the-deep/deep-server-utils",

    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Text Processing :: General",
        "Topic :: Utilities",
        "Operating System :: OS Independent",
    ],

    packages=find_namespace_packages(),
    install_requires=[
        'boto3==1.9.88',
        'elasticsearch==7.8.1',
        'requests-aws4auth==1.0',
    ],
    zip_safe=False
)



import string
import json
import langdetect
from tqdm import tqdm

from lead.models import LeadPreview

# Translate punctuations to spaces
PUNC_TRANS_TABLE = str.maketrans({x: ' ' for x in string.punctuation})

ENGLISH = 'en'
FRENCH = 'fr'
SPANISH = 'es'

DOCUMENT_STATS: dict = {
    ENGLISH: dict(count=0, corpus_length=0),
    FRENCH: dict(count=0, corpus_length=0),
    SPANISH: dict(count=0, corpus_length=0),
}

TRIGRAMS_COUNT: dict = {
    ENGLISH: {},
    FRENCH: {},
    SPANISH: {}
}


def remove_puncs_and_extra_spaces(text: str) -> str:
    return ' '.join(text.translate(PUNC_TRANS_TABLE).split()).lower()


def update_trigrams_count(text: str, lang: str = ENGLISH) -> None:
    for index in range(len(text) - 2):
        trigram = text[index:index + 3]
        TRIGRAMS_COUNT[lang][trigram] = TRIGRAMS_COUNT[lang].get(trigram, 0) + 1
    return


def get_total_preview_count():
    text_previews = LeadPreview.objects.all().filter(text_extract__isnull=False)
    return text_previews.count()


def generate_trigrams(json_path, limit=None):
    """
    Generates trigrams from LeadPreview Objects and writes them to a json file.
    JSON will have the following structure:
    ```
    {
        "en": { <trigram>: <trigram_count>, ... },
        "es": { <trigram>: <trigram_count>, ... },
        "fr": { <trigram>: <trigram_count>, ... }
    }
    ```
    returns: document counts and sizes dictionary
    ```
    {
        "en": { "count": <english docs count>, "corpus_length": <all_documents_byte_size> },
        "es": { "count": <spanish docs count>, "corpus_length": <all_documents_byte_size> },
        "fr": { "count": <french docs count>, "corpus_length": <all_documents_byte_size> },
    }
    ```
    """
    text_previews = LeadPreview.objects.all().filter(text_extract__isnull=False)
    if limit:
        text_previews = text_previews[:limit]
    total_count = text_previews.count()
    print(f'TOTAL previews: {total_count}')

    if json_path:
        json_file = open(json_path, 'w')

    obj_iterator = text_previews.values_list('text_extract', flat=True).iterator(chunk_size=500)
    for preview in tqdm(obj_iterator, total=total_count):
        if not preview or len(preview.strip()) == 0:
            continue
        try:
            language = langdetect.detect(preview)
        except Exception:
            continue
        if language not in [ENGLISH, FRENCH, SPANISH]:
            continue
        cleaned = remove_puncs_and_extra_spaces(preview)
        cleaned_len = len(cleaned)
        if not cleaned or cleaned_len < 3:
            continue

        # UPdate document STATS
        DOCUMENT_STATS[language]['count'] += 1
        DOCUMENT_STATS[language]['corpus_length'] += cleaned_len
        update_trigrams_count(cleaned, language)
    print(f'Now writing to {json_path}')
    if json_path:
        json.dump(TRIGRAMS_COUNT, json_file, indent=2)
    print('All Done!!!')
    print(DOCUMENT_STATS)
    return DOCUMENT_STATS


if __name__ == '__main__':
    generate_trigrams('raw_trigrams.json')



from typing import Optional
import string

from requests_aws4auth import AWS4Auth
import boto3
from elasticsearch import Elasticsearch as ES, RequestsHttpConnection

PUNC_TRANS_TABLE = str.maketrans({x: ' ' for x in string.punctuation})


def es_wrapper(endpoint: str, region: str = 'us-east-1', profile_name: Optional[str] = None):
    """
    Wrapper function for Elasticsearch
    """
    credentials = boto3.Session(profile_name=profile_name).get_credentials()
    awsauth = AWS4Auth(
        credentials.access_key,
        credentials.secret_key,
        region,
        'es',
        session_token=credentials.token
    )

    return ES(
        endpoint,
        http_auth=awsauth,
        scheme="https",
        use_ssl=False,
        verify_certs=True,
        connection_class=RequestsHttpConnection
    )


def remove_puncs_and_extra_spaces(text: str) -> str:
    return ' '.join(text.translate(PUNC_TRANS_TABLE).split()).lower()
